/*
 * a.lp_mp3 - Open Source Atmel AVR based MP3 Player
 * Copyright (c) 2003-2004 K. John '2B|!2B' Crispin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 *
 * Feedback, Bugs.... mail john{AT}phrozen.org
 *
 */ 
 
 // TODO: 
 // 	improve poerdown as mmc and vs dont actually powerdown, just the uC does

// avr includes
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

// my lib header
#include "types.h"
#include "USART.h"
#include "spi.h"
#include "delay.h"
#include "vs1001.h"
#include "mmc.h"
#include "fat.h"

// defines for the user buttons
#define BUTTONVOL_PORT 		PORTA
#define BUTTONVOL_DDR 		DDRA
#define BUTTONVOL_PIN 		PINA
#define BUTTONCTL_PORT 		PORTA
#define BUTTONCTL_DDR 		DDRA
#define BUTTONCTL_PIN 		PINA

#define BTN_VOL_DOWN		PA0
#define BTN_VOL_UP			PA1
#define BTN_SONG_DOWN		PA2
#define BTN_SONG_UP			PA3
#define BTN_PLAY			PA4

// defines for leds
#define LED_PORT   		PORTA
#define LED_DDR 		DDRA
#define LED_PLAY_PIN	PA6
#define PLAY_LED_ON	cbi(LED_PORT, LED_PLAY_PIN)
#define PLAY_LED_OFF	sbi(LED_PORT, LED_PLAY_PIN)

#define LED_START_PIN	PA7
#define START_LED_ON	cbi(LED_PORT, LED_START_PIN)
#define START_LED_OFF	sbi(LED_PORT, LED_START_PIN)

// defines for the powerdown button
#define INT_PORT 		PORTD
#define INT_PIN 		PIND
#define INT_DDR 		DDRD
#define POWER_INT_PIN 	PD3

// our beeps
#define BEEP(x)	{VS1001_volume(0x30, 0x30); \
				beep(0x30, 70, x);	\
				VS1001_volume(player_volume, player_volume); \
				}

// player states
#define IDLE 				0
#define PLAYING_MMC  		2
#define PLAYING_MMC_INIT  	6
#define SHUTDOWN			7

// Play modes
#define PS_SEQUENCE 0x01	// plays files in the order found. then the folders
#define PS_PLAYLIST	0x02	// the player found a playlist file on the mmc

// what is the player doing atm ? and how is it doing it ?
typedef struct _P_STATE{
	// the current state
	u08 state;
	// what play mode are we using ?
	u08 mode;
	// what song are we on?
	u08 file_index;
}PLAYER_STATE;

// holds the last value read from the button port
static u08 last_button = 0x00;
// the current volume of the player
static u08 player_volume = 0x00;
// used to remember if we were just in powerrdown.
static u08 allow_powerdown = 0x01;
// the powerdown counter so that you have to hold the button for a while before the plyaer shuts down
static u08 powerdown_counter = 0x00;
// the number of beeps to generate
static u08 num_beeps = 0;
// the last byte that we received
static u08 u_static = 0x00;
// what are we doing atm ?
static PLAYER_STATE ps;
// our mp3 file ptr
static FILE fp;

// BEEP
void beep(u08 freq, u16 length, u08 intervals){
	VS1001_SW_reset();
	VS1001_send_zeros(32);
	do{
		VS1001_sine(0x01, freq);
		delay_ms(length);
		VS1001_sine(0x00, 0x00);
		delay_ms(length);
	}while(--intervals);
	
};

// gets called during powerdown to re-enable the player
SIGNAL(SIG_INTERRUPT1){
	// disable int1
	cbi(GICR, INT1);	
};


// this function shutsdown the mmc and the vs1001 and the starts to power down the cpu
void powerdown(void){
	// shutdown MMC
	// shutdown vs1001
	// setup powerdown
	cbi(MCUCR, SM0);
	sbi(MCUCR, SM1);
	cbi(MCUCR, SM2);
	// enable powerdown
	sbi(MCUCR, SE);
	do{
		// enable int1
		sbi(GICR, INT1);	
		// do the actual powerdown
		SLEEP;
	} while(bit_is_set(BUTTONVOL_PIN, BTN_VOL_DOWN));
	// disable powerdown
	cbi(MCUCR, SE);
	// disable int1
	cbi(GICR, INT1);	
	// restart MMC
	// restart vs1001
};

// this function sets up the interrupt and ports used by the buttons
void setup_buttons(void){
	// pull the port down so the interrupt gets triggered properly
	cbi(BUTTONVOL_DDR, BTN_VOL_UP);
	cbi(BUTTONVOL_DDR, BTN_VOL_DOWN);
	cbi(BUTTONCTL_DDR, BTN_SONG_UP);
	cbi(BUTTONCTL_DDR, BTN_SONG_DOWN);
	cbi(BUTTONCTL_DDR, BTN_PLAY);
	
	cbi(BUTTONVOL_PORT, BTN_VOL_UP);
	cbi(BUTTONVOL_PORT, BTN_VOL_DOWN);
	cbi(BUTTONCTL_PORT, BTN_SONG_UP);
	cbi(BUTTONCTL_PORT, BTN_SONG_DOWN);
	cbi(BUTTONCTL_PORT, BTN_PLAY);
	// setup the leds
	sbi(LED_DDR, LED_PLAY_PIN);
	PLAY_LED_OFF;
	sbi(LED_DDR, LED_START_PIN);
	START_LED_ON;
	
	// setup the powerdown button -> int1
	cbi(INT_PORT, POWER_INT_PIN);// setup the powerdown button -> int1
	cbi(INT_DDR, POWER_INT_PIN);// setup the powerdown button -> int1
	
};



// N.C.
void print_help(void){
	u08 i=0x12;
	for(;i<0x1d; i++) USART_sendpgmstring(i,0x01);
	USART_newline();
};

// play the next file
void play_next(void){
	ps.state = PLAYING_MMC_INIT;
	fclose(&fp);
	num_beeps = 1;
};

// go back a file
void play_prev(void){
	ps.file_index--;
	if(ps.file_index > 0) ps.file_index--;
	// start player init process
	ps.state = PLAYING_MMC_INIT;
	fclose(&fp);
	num_beeps = 2;
};

// turn up volume
void vol_up(void){
	// check if we can go any louder
	if(player_volume > 1) player_volume -= 2;
	// set new volume
	VS1001_volume(player_volume, player_volume);	
};

// turn down volume
void vol_down(void){
	// check if we can go any quiter
	if(player_volume < 180) player_volume += 2;
	// set new volume
	VS1001_volume(player_volume, player_volume);	
};

// this function processes the button messages
void process_buttons(void){
	// find out which buttons were pressed
	u08	tmp  = 	((bit_is_set(BUTTONVOL_PIN, BTN_VOL_UP) != 0) << 4) +
				((bit_is_set(BUTTONVOL_PIN, BTN_VOL_DOWN) != 0) << 3) +
				((bit_is_set(BUTTONCTL_PIN, BTN_SONG_UP) != 0) << 2) +
				((bit_is_set(BUTTONCTL_PIN, BTN_SONG_DOWN) != 0) << 1) +
				((bit_is_set(BUTTONCTL_PIN, BTN_PLAY) != 0) << 0);
	
	//u08 tmp = inp(BUTTON_PIN);
	// invert as 0v -> button pressed
	tmp = ~tmp;
	// make sure the button was not pressed before 
	u08 cleaned_val = tmp ^ last_button;
	cleaned_val = cleaned_val & tmp;
	// store current val as last val
	last_button = tmp;
	
	// button 1 was pressed -> song back
	if(cleaned_val&(1<<1)) play_prev();	
	// button 2 was pressed -> song next
	if(cleaned_val&(1<<2))	play_next();
	// button 3 was pressed -> vol -
	if(cleaned_val&(1<<3)) vol_down();
	// button 4 was pressed -> vol +
	if(cleaned_val&(1<<4)) vol_up();
	// button 5 was pressed -> play/stop
	if(cleaned_val&(1<<0)){
		if((ps.state == PLAYING_MMC_INIT) || (ps.state == PLAYING_MMC)){
			// stop the player
			ps.state = IDLE;
			PLAY_LED_OFF;
			BEEP(3);
		} else {
			// go to start of file
			ps.state = PLAYING_MMC_INIT;
			num_beeps = 1;
		};
	};	
	// button 6 was pressed -> soft off
	if(bit_is_clear(INT_PIN, POWER_INT_PIN)){
		if(!allow_powerdown) return;
		powerdown_counter++;
		if(powerdown_counter < 80){
			if(ps.state == IDLE) delay_ms(50);
			return;
		};
		START_LED_ON;
		PLAY_LED_ON;
		//USART_sendstring("Shutting Down...", 0x01);
		USART_sendpgmstring(0x2c,0x01);
		VS1001_SW_reset();
		BEEP(4);
		allow_powerdown = 0x00;
		while(bit_is_clear(INT_PIN, POWER_INT_PIN)){};
		powerdown();
		BEEP(4);
		START_LED_OFF;
		PLAY_LED_OFF;
		//USART_sendstring("Powering Up...", 0x01);
		USART_sendpgmstring(0x2d,0x01);
		ps.state = IDLE;
	} else {
		allow_powerdown = 0x1;
		powerdown_counter = 0x00;
	};
};

// gets called by usart.c
void USART_callback(unsigned char u){
	u_static = u;
};

// gets called once for every 'main loop' loop
void process_usart(void){
	if(u_static == 0x00) return;
	USART_send(u_static);
	USART_newline();
	
	if(u_static == 'h') print_help();
	if(u_static == '0') VS1001_sine(0x00,0x00);
	if(u_static == '1') VS1001_sine(0x01,0x30);
	if(u_static == '2') VS1001_print_registers();
	if(u_static == '3') VS1001_print_audio_data();
	if(u_static == '4') USART_sendint(VS1001_get_decode_time());
	if(u_static == '5') BEEP(5);
	if(u_static == '6') ps.state = PLAYING_MMC_INIT;
	if(u_static == ','){
		play_prev();
	};	 
	if(u_static == '.'){
		play_next();
	};
	if(u_static == '+'){
		vol_up();
	};
	if(u_static == '-'){
		vol_down();
	};
	if(u_static == 's'){
		ps.state = IDLE;
		fclose(&fp);
	};
	
	u_static = 0x00;
};


int main (void){	
    // setup the player 
	ps.state = IDLE;
	ps.mode = PS_SEQUENCE;
	ps.file_index = 0;
	// make sure the ptr for the mp3 file is setup correctly
	fclose(&fp);
	setup_buttons();
	
	// setup USART for serial communications
	// for 8mhz internal resonator
	//USART_init(204);//USART_BAUD(2400));
	//USART_init(101);//USART_BAUD(4800));
	USART_init(50);//USART_BAUD(9600));
	//USART_init(24);//USART_BAUD(19200));
	//USART_init(11);//USART_BAUD(38400));
	//USART_init(7);//USART_BAUD(57600));
	USART_clear_screen();
	
	//USART_sendstring("starting ...", 0x01);
	USART_sendpgmstring(0x00,0x01);
	
	// give mmc time
	delay_ms(1000);
	// set the MMC pins
	MMC_hw_init();
	spi_init();
	sei();
	USART_newline();
	//USART_sendstring("starting VS1001 ...", 0x01);
	USART_sendpgmstring(0x01,0x01);
	VS1001_init();
    VS1001_volume(player_volume, player_volume);	
	//USART_sendstring("VS1001 running ...", 0x01);
	USART_sendpgmstring(0x02,0x01);
	USART_newline();
	//USART_sendstring("starting MMC ...", 0x01);	
	USART_sendpgmstring(0x03,0x01);
	// setup mmc
	MMC_init();
	//USART_sendstring("MMC OK...", 0x01);	
	USART_sendpgmstring(0x04,0x01);
	USART_newline();
	//USART_sendstring("starting FAT...", 0x01);	
	USART_sendpgmstring(0x05,0x01);
	// setup fat layer
	FAT_startup();
	//USART_sendstring("FAT OK...", 0x01);	
	USART_sendpgmstring(0x06,0x01);
	USART_newline();
	// print help
	print_help();
	BEEP(3);
	// flash the led 
	
	PLAY_LED_OFF;
	START_LED_OFF;
	
	
	// ------id3 example-----
	
	// open second file
	fopenc(1, 'r',NULL, &fp);
	// print id3 tag
	USART_sendstring("this can take some time...", 0x01);
	fid3tag(&fp);
	// close file
	fclose(&fp);
	
	
	
	// ------file example-----
	/*
	// open the first file in the first folder on the first partition
	fopenc(0, 'r', NULL, &fp);
	// seek ~83k into the file
	if(!fseek(&fp, 0x01, 0x4960,SEEK_CUR)){
			USART_sendstring("seek_error", 0x01);
	} else {
		// get 32 bytes and print them
		if(fget(&fp,0x20)){
			USART_sendstring("-----", 0x01);
			USART_sendstring(fio_buf, 0x01);
			int k;
			for(k= 0;k < 0x20; k++) USART_sendhex(fio_buf[k]);
			USART_newline();
			USART_sendstring("-----", 0x01);
		} else {
			USART_sendstring("get_error", 0x01);
		};
	};
	fclose(&fp);
	*/
	
	// ------dir example------
	/*
	// do some directory stuff
	dir_list(NULL);
	dir_enter("NEUERO~1", NULL);
	dir_list(NULL);
	dir_enter("ABC     ", NULL);
	dir_list(NULL);
	dir_leave(NULL);
	dir_list(NULL);
	dir_leave(NULL);
	dir_list(NULL);
	*/
	
	// ------mp3 example----
	/*
	// open the 3 mp3 file on the media
	fopenc(1, 'r',"*       MP3", &fp);
	// play mp3
	while(fplay_sector(&fp)){};
	// close file
	fclose(&fp);
	*/
	
	// print root folder
	dir_list(NULL);
	// put player in idle state
	ps.state = IDLE;
	num_beeps = 0;
	while(1){
		// has the user pressed a button ?
		process_buttons();
		// did we get data on the usart?
		process_usart();
		// what state are we in ?.
		switch(ps.state){
				// restart
			case PLAYING_MMC_INIT:
			// reset. decoder
				VS1001_SW_reset();
				// generate beep
				BEEP(num_beeps);
				delay_ms(100);
				//USART_sendstring("Starting MMC stream", 0x01);
				USART_sendpgmstring(0x07,0x01);
				// player led on
				PLAY_LED_ON;
				// find the next song
				if(fopenc(ps.file_index,'r', "*       MP3", &fp)){
					ps.state = PLAYING_MMC;
					//USART_sendstring("__FOUND", 0x01);
					USART_sendpgmstring(0x09,0x01);
					ps.file_index++;
				} else {
					ps.file_index = 0;
					//USART_sendstring("Last File Found", 0x01);				
					USART_sendpgmstring(0x0a,0x01);
					//BEEP(6);
					break;
				};
			case PLAYING_MMC:
				// play the next sector and restart if it is the last ne
				if(!fplay_sector(&fp)){
					fclose(&fp);
					ps.state = PLAYING_MMC_INIT;
					num_beeps = 1;
					//USART_sendstring("file end reached",0x01);
					USART_sendpgmstring(0x08,0x01);
				};
				break;
			default:
				break;
		};
	};
	
	return 0;
}




















