/*
 * a.lp_mp3 - Open Source Atmel AVR based MP3 Player
 * Copyright (c) 2003-2004 K. John '2B|!2B' Crispin
 *
 * For Private Use :
 * -----------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 *
 */ 

 /* Feedback, Bugs, Donations.... mail john{AT}phrozen.org */
 /* particularly proud of this one ;) */
 /* this is only the 8k uC version */
 
 /* 16k uC version coming soon with the following new features 
 ||
 ||	TODO :
 ||		add fgetc	
 ||		add fgets
 ||		add support for long filename
 ||     major improvement of fseek as it it slow as hell
 ||     add write support
 ||		add fput
 ||		add fputs
 ||		add a PARTITION struct and integrate it with the existing code
 ||		add support for more than one partition
 ||		add part_chg
 ||		add fcopy
 ||     add fdel
 ||		add frename
 ||		add dir_mk
 ||		add dir_rm
 ||		add dir_rename
 ||		
 ||	
 */
 

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

#include "types.h"
#include "USART.h"
#include "mmc.h"
#include "vs1001.h"
#include "spi.h"
#include "fat.h" 

// starts by scanning the BPB
// sector needs to be equal to the first sector of the first partition
// use fat_get_part_start to get this value
//
// the first fat32 sector is set up as follows -->
//	u08		jmpboot[3]		0xEB 0x?? 0x90		-- instructions to jump to the boot loader		
//	u08		OEMName[8]		-- identifies system that formatted the card
//	u08		BPB[42]			-- BIOS Parameter Block -- see below
//	.
//	.
//	.
//	u08		0xAA			-- signature
//	u08		0x55			-- signature
//
// fat32 BPB -->
//	u08 bytes_per_sec[2]		-- the size of a sector. for MMC this is always 512
//	u08 sec_per_clus	        -- the number of sectrs per cluster. for MMC normally 1
//	u08 rsvd_sec[2]				-- reserved sectors for additional boot code. fat32 normally has 32 sectors
//	u08 num_fats	           	-- number of fat-chains on the disk ? leave at 2 to be sure
//	u08 root_dir_ent[2]     	-- 0 for fat32. used by fat16
//	u08 tot_sec_16[2]         	-- 0 for fat32. used by fat16
//	u08 media_type             	-- used by OS to determine removeable, fixed ... used by msdos 1.x
//	u08 sec_per_fat[2]        	-- 0 for fat32. used by fat16
//	u08 sec_per_track[2]      	-- something for geometry ?
//	u08 num_heads[2]            -- number of heads. 
//	u08 hidden_secs[4]			-- something for HW int 0x13
//	u08 tot_sec_32[4]			-- the number of sectors on this partition
//	u08 sec_per_fat_32[4]      	-- sectors per fat
//	u08 ext_flags[2]			-- extended partition ?
//	u08 fs_ver[2]		        -- ?? something for fs
//	u08 root_clus[4]	       	-- offset to the sector that contains the root directory
//	u08 fs_inf[2]	          	-- ?? something for fs
//	u08 bpb_bak[2]          	-- backup of bpb in reserved area of media



// defines for the offsets of the values in the bpb that we are interested in

#define FAT_OEM_START		    3
#define FAT_OEM_STOP			10

#define FAT_BYTES_PER_SECTOR	11
#define FAT_SEC_PER_CLUS		13
#define FAT_RSVD_SEC_CNT		14
#define FAT_NUM_FATS			16
#define FAT_ROOT_ENT_CNT		17
#define FAT_TOTSEC				32
#define FAT_SIZE				36
#define FAT_ROOT_CLUS			44

// file attribs
#define FAT_READ_ONLY			0x01
#define FAT_HIDDEN				0x02
#define FAT_SYS					0x04
#define FAT_VOL_ID				0x08
#define FAT_DIRECTORY			0x10
#define FAT_ARCHIVE				0x20
#define FAT_LONG_NAME  			0x0F

static u16 bytes_per_sec = 0;
static u08 sec_per_clus = 0;
static u16 rsvd_sec_cnt = 0;
static u08 num_fats = 0;
static u16 tot_sech = 0;
static u16 tot_secl = 0;
static u16 root_clush = 0;
static u16 root_clusl = 0;
static u16 fat_sizeh = 0;
static u16 fat_sizel = 0;

// our offsets
static u16 first_part_sector = 0;
static u16 first_fat_sector = 0;
static u16 first_data_sectorl = 0;

// name of our volume
// gets set about a milion times
// this is a definate todo
static u08 volume_id[8];

// the directiory that gets used if no other is specified
static DIR directory;



/*------------------ FAT LAYER --------------*/



void FAT_print_sector(u16 sectorh, u16 sectorl){
	// bytes to read
	u16 length = 512;
	// print sector 
	USART_sendhex(sectorh>>8);
	USART_sendhex(sectorh);
	USART_sendhex(sectorl>>8);
	USART_sendhex(sectorl);
	// turn sectors into byte addr
	u16 myh = (sectorh << 9) + (sectorl >> 7);
	u16 myl = sectorl << 9;
	USART_newline();
	
	// start reading sector
	MMC_get_sec_start(sectorh, sectorl);
	// read bytes and print them
	while(length){
		if((length%16) == 0) {
			USART_sendhex(myh>>8);
			USART_sendhex(myh);
			USART_sendhex(myl>>8);
			USART_sendhex(myl);
			USART_send('|');
			USART_send(' ');
			myl += 0x10;
		};
		u08 tmp = MMC_get_sec_next();
		USART_sendhex(tmp);
		length--;
		// new line every 16 bytes
		if((length%16) == 0) USART_newline();
	};
	// stop reading sector
	MMC_get_sec_stop();
};


// reads the mbr
// check for fat32 identifier
// finds the start of the first partition
u08 FAT_read_mbr(void){
	// the number of bytes read
	u16 length = 0;	
	// start reading the mbr
	MMC_get_sec_start(0x00,0x00);
	// get the data
	while(length < 512){
		length++;
		// get next byte
		u08 tmp = MMC_get_sec_next();
		// make sure this is a fat 32 partition
		if(length == 451) {
			USART_sendhex(tmp);
		};
		// get the start sector of the first partition
		if(length == 455) first_part_sector = tmp;
		if(length == 456) first_part_sector += ((u16)tmp) << 8;
	};
	// stop reading from the mmc
	MMC_get_sec_stop();
	// FAB
	return FAT_OK;
};
// scans the BPB in the first(0x00) sector of the first partition
void FAT_read_first_part_sector(void){
	// the number of bytes read
	u16 length = 0;	
	// start reading from the mmc
	MMC_get_sec_start(0x00, first_part_sector);
	while(length < 512){
		// get the next byte
		u08 tmp = MMC_get_sec_next();
		// get the id string
		if((length >= FAT_OEM_START) &&( length <= FAT_OEM_STOP)){
			USART_send(tmp);
		};
		// read fat data
		switch(length){
		case FAT_BYTES_PER_SECTOR:
			// get upper byte;
			bytes_per_sec = MMC_get_sec_next();
			length++;
			// assemble word
			bytes_per_sec = (bytes_per_sec << 8) + tmp;
			break;
		case FAT_SEC_PER_CLUS:
			sec_per_clus = tmp;
			break;
		case FAT_RSVD_SEC_CNT:
			// get upper byte;
			rsvd_sec_cnt = MMC_get_sec_next();
			length++;
			// assemble word
			rsvd_sec_cnt = (rsvd_sec_cnt << 8) + tmp;
			break;
		case FAT_NUM_FATS:	
			num_fats = tmp;
			break;
		case FAT_TOTSEC:
			// get second byte;
			tot_secl = MMC_get_sec_next();
			length++;
			// assemble word
			tot_secl = (tot_secl << 8) + tmp;
			
			// get byte 3 and 4
			tmp = MMC_get_sec_next();
			length++;
			// get second byte;
			tot_sech = MMC_get_sec_next();
			length++;
			// assemble word
			tot_sech = (tot_sech << 8) + tmp;
			break;
		case FAT_SIZE:
			// get second byte;
			fat_sizel = MMC_get_sec_next();
			length++;
			// assemble word
			fat_sizel = (fat_sizel << 8) + tmp;
			
			// get byte 3 and 4
			tmp = MMC_get_sec_next();
			length++;
			// get second byte;
			fat_sizeh = MMC_get_sec_next();
			length++;
			// assemble word
			fat_sizeh = (fat_sizeh << 8) + tmp;
			break;
		case FAT_ROOT_CLUS:
			// get second byte;
			root_clusl = MMC_get_sec_next();
			length++;
			// assemble word
			root_clusl = (root_clusl << 8) + tmp;
			
			// get byte 3 and 4
			tmp = MMC_get_sec_next();
			length++;
			// get second byte;
			root_clush = MMC_get_sec_next();
			length++;
			// assemble word
			root_clush = (root_clush << 8) + tmp;
			break;
			
		};
		length++;
	};
	MMC_get_sec_stop();
};

// this function gets the addr in the data region for the next cluster of a file
void FAT_get_next_clus_addr(u16* sectorh, u16* sectorl){
	// determine the sector that we need to read and which dword defines our  next cluster address
	u16 myh = *sectorh;
	u16 myl = *sectorl;
	u16 myoffset = *sectorl;
	// work out the mod
	while(myoffset > 127) myoffset -= 128;
	// turn sectors into byte address
	myl = (myl >> 7) + ((myh & 0x7f) << 9);
	myh = myh >> 7;
	myoffset += 2;
	if( myoffset >127){
		myoffset -= 128;
		myl++;
		if(myl == 0x0) myh++;
	};
	myl += first_fat_sector;
	if(myl < first_fat_sector) myh++;
	// get the data
	MMC_get_part_sec_start(myh, myl,myoffset * 4,4);
	*sectorl = MMC_get_sec_next();
	*sectorl += (MMC_get_sec_next() << 8);
	*sectorh = MMC_get_sec_next();
	*sectorh += (MMC_get_sec_next() << 8);
	*sectorl -= 2;
	if( *sectorl > 0xfffd) sectorh--;
	// fine
	MMC_get_sec_stop();
};




void fio_init(void);

// 
// 1.) reads the MBR
// 	   gets start sector of first partition
// 2.) verifies start sector
// 3.) get BPB
// 4.) find fat start
// 5.) start reading fat


void FAT_startup(void){
	// 1.) read MBR
	u08 err = FAT_read_mbr();
	
	// check for invalid fs
	if(err == FAT_INVALID_ID){
		//USART_sendstring("invalid fs",0x01);
		USART_sendpgmstring(0x21,0x01);
		return;
	};
	
	// 2.) any other errors --> todo
	if(err != FAT_OK){
		//USART_sendstring("invalid fs",0x01);
		USART_sendpgmstring(0x21,0x01);
		return;
	};
	
	
	// 3.) read BPB
	FAT_read_first_part_sector();
	
	// 4.) find fat start
	first_fat_sector = first_part_sector + rsvd_sec_cnt;
	first_data_sectorl = first_fat_sector + (num_fats * fat_sizel);
	
	u08 i;
	for(i=0; i<8; i++){
		volume_id[i] = 'X';
	};	
	USART_newline();
	
	/*
	// print out info for debug
	USART_sendhex(first_part_sector>>8);
	USART_sendhex(first_part_sector);
	USART_newline();
	
	USART_sendhex(bytes_per_sec>>8);
	USART_sendhex(bytes_per_sec);
	USART_newline();
	
	USART_sendhex(sec_per_clus);
	USART_newline();
	
	USART_sendhex(rsvd_sec_cnt>>8);
	USART_sendhex(rsvd_sec_cnt);
	USART_newline();
	
	USART_sendhex(num_fats);
	USART_newline();
	
	USART_sendhex(tot_sech>>8);
	USART_sendhex(tot_sech);
	USART_newline();
	
	USART_sendhex(tot_secl>>8);
	USART_sendhex(tot_secl);
	USART_newline();
	
	USART_sendhex(root_clusl>>8);
	USART_sendhex(root_clusl);
	USART_newline();
	
	USART_sendhex(first_data_sectorl>>8);
	USART_sendhex(first_data_sectorl);
	USART_newline();
	
	USART_sendhex(first_fat_sector>>8);
	USART_sendhex(first_fat_sector);
	USART_newline();
	
	USART_sendhex(fat_sizel>>8);
	USART_sendhex(fat_sizel);
	USART_newline();
	*/
	fio_init();
	return;
};



/*------------------ FIO LAYER --------------*/



// compares a 11 byte block from the fat table with a 8 byte name and 3 byte extension. 
// '*' is supported as a wildcard
bool fio_cmp_names(u08 *from_fat, u08 *name, u08 *extension){
	u08 count = 0;
	for(;count < 8; count++){
		if(name[count] == '*'){
			count = 100;
		} else {
			if(name[count] != from_fat[count]) 
				return false;
		};
	};
	
	if(extension){
		for(;count < 3; count++){
			if(extension[count] =='*'){
				count = 100;
			} else {
				if(extension[count] != from_fat[8+count]) 
					return false;
			};
		};
	};
	return true;
};

// setup all the data needed by file i/o. gets called by FAT_startup
void fio_init(void){
	// on a ms compliant fat32 the 0 sector is always the first sector of the root directory
	directory.dir_currenth = 0;
	directory.dir_currentl = 0;
};

// takes a 32 byte long fat entry and opens the file contained within
void fio_load_file_ptr(u08 *fat_entry, FILE *stream){
	u08 i;
	stream->file_first_sectorl = (((u16)fat_entry[27])<<8) + fat_entry[26];
	stream->file_first_sectorh = (((u16)fat_entry[21])<<8) + fat_entry[20];
	
	// substract 2 because the address in the fat is always off by two. this is what ms wants ?!?
	stream->file_first_sectorl -= 2;
	if(stream->file_first_sectorl > 0xfffd){
		stream->file_first_sectorh--;
	};
	
	stream->file_current_sectorl = stream->file_first_sectorl;
	stream->file_current_sectorh = stream->file_first_sectorh;
	
	stream->file_full_sizel = stream->file_sizel = (((u16)fat_entry[29])<<8) + fat_entry[28];
	stream->file_full_sizeh = stream->file_sizeh = (((u16)fat_entry[31])<<8) + fat_entry[30];
	
	stream->file_offset = 0x00;
	
	// USART_sendstring("File",0x01);
	USART_sendpgmstring(0x24,0x00);
	
	for(i=0; i<11; i++){
		stream->name[i] = fat_entry[i];
		USART_send(fat_entry[i]);
	};
	USART_newline();
};

// takes a 32 byte long fat entry and opens the dir contained within
void fio_load_dir_ptr(u08 *fat_entry, DIR *stream){
	// get the dirposition
	stream->dir_currentl = (((u16)fat_entry[27])<<8) + fat_entry[26];
	stream->dir_currenth = (((u16)fat_entry[21])<<8) + fat_entry[20];
	// substract 2 because the address in the fat is always two wrong. this is what ms wants ?!?
	if(!((stream->dir_currentl==0x0)&&(stream->dir_currenth==0x0))){
		stream->dir_currentl -= 2;
		if(stream->dir_currentl > 0xfffd){
			stream->dir_currenth--;
		};
	};
};


// gets part of a file
// this function is dumb and does not check for errors.
// dont use on its own
// used by fget & co.
void fio_get_file_section(FILE *stream, u16 size, u08 *data){
	// determine the sector that we need to read and which dword defines our  next cluster address
	u16 myh = stream->file_current_sectorh;
	u16 myl = stream->file_current_sectorl;
	myl += first_data_sectorl;
	if(myl < first_data_sectorl) myh++;
	// get the data
	MMC_get_part_sec_start(myh, myl,stream->file_offset,size);
	u08 i=0;
	for(;i<size;i++) data[i] = MMC_get_sec_next();
	MMC_get_sec_stop();
	// fine
};

// used to get info from the current directoy
// modes :
//	FIO_PRINT - 0x01 - print directory
//  FIO_FILENAME - 0x02 - find file by name
//  FIO_DIRNAME - 0x03 - find dir by name
//  FIO_FILENUM - 0x04 - find file by number 0 based
//  FIO_DIRNUM - 0x05 - find dir by number 0 based
bool fio_dir_ctl(u08 mode, FILE *stream, DIR *dir, u08* name, u08 count){
//void fio_print_dir(void){
	// the sector that we are on
	u16 myhfat;
	u16 mylfat;
	u16 myh;
	u16 myl;
	// how many files/dirs were foud
	u08 index = 0;
	// is there data worth looking for in another sector ?
	u08 loop = 1;
	// how many entries were read from this sector
	u08 len = 0;
	
	if(mode == FIO_PRINT){
		// print the table header
		USART_sendpgmstring(0x23,0x00);
		USART_sendpgmstring(0x22,0x00);
		USART_sendpgmstring(0x1f,0x00);
		USART_sendpgmstring(0x20,0x01);	
	};
	
	myhfat = directory.dir_currenth;
	mylfat = directory.dir_currentl;
	
	do{
		myl = mylfat;
		myh = myhfat;

		// add Data section offset
		myl += first_data_sectorl;
		if(myl < first_data_sectorl) myh++;
		
		// start getting the data from the mmc
		MMC_get_sec_start(myh, myl);
		// there are 16 directory entries in each directory sector
		// checkout the msdoc of fat for info
		len = 0;
		while (len < 16){
			u08 tmp[32];
			u08 i;
			// get next entry
			for(i =0; i< 32; i++){
				tmp[i] = MMC_get_sec_next();
			};
			if(tmp[0] == 0x00) loop = 0;
			// make sure that there is an entry and that we want to process it and that it is valid
			if((tmp[0] >= 0x20) && (tmp[0] != 0xE5)){
				// do want to print ?
				if ((tmp[11]&FAT_LONG_NAME) == FAT_LONG_NAME){
					// check if this is a long dir entry
					if(mode == FIO_PRINT){
						//for(i=0; i<11; i++){
						//	if(tmp[i]>0xf) USART_send(tmp[i]);
						//};
						//USART_newline();
					} 
				}
				// check if it is a file
				else if(((tmp[11]&(FAT_DIRECTORY|FAT_VOL_ID)) == 0x00)){
					// do want to print ?
					if(mode == FIO_PRINT){
						// spaces added after filename
						u08 spaces = 3;
						// USART_sendstring("File",0x01);
						USART_sendpgmstring(0x24,0x00);
						// print the filename
						for(i=0; i<11; i++){
								if(i==8) USART_send('.');
								if(tmp[i] == ' ')
									spaces++;
								else
									USART_send(tmp[i]);
						};
						USART_sendspaces(spaces);
						// get the file size
						myl = (((u16)tmp[29])<<8) + tmp[28];
						myh = (((u16)tmp[31])<<8) + tmp[30];
						USART_sendhex(myh>>8);
						USART_sendhex(myh);
						USART_sendhex(myl>>8);
						USART_sendhex(myl);
						USART_sendspaces(3);
						// get the fileposition
						myl = (((u16)tmp[27])<<8) + tmp[26];
						myh = (((u16)tmp[21])<<8) + tmp[20];
						// substract 2 because the address in the fat is always off by two. this is what ms wants ?!?
						myl -= 2;
						if(myl > 0xfffd){
							myh--;
						};
						USART_sendhex(myh>>8);
						USART_sendhex(myh);
						USART_sendhex(myl>>8);
						USART_sendhex(myl);
						USART_newline();
					} 
					// are we looking for a file by number ?
					else if((mode == FIO_FILENUM)&&(loop)) {
						// is this the file we are looking for ?
						if(index == count){
							// check if we are supposed to use a mask
							if(name) {
								USART_sendstring("cmp : ", 0x00);
								USART_send_mem(name, 11);
								USART_sendstring(" & ", 0x00);
								USART_send_mem(tmp, 11);
								USART_newline();
								if(fio_cmp_names(tmp, name, &name[8])){
									fio_load_file_ptr(tmp, stream);
									loop = 0;
								};
								//index--;
							} else {
								fio_load_file_ptr(tmp, stream);
								loop = 0;
							};
						} else {
							if(!name){
								index++;
							} else {
								if(fio_cmp_names(tmp, name, &name[8])){
									index++;
								};
							}
						};						
					} else if(mode == FIO_FILENAME){
						if(fio_cmp_names(tmp, name, &name[8])){
							fio_load_file_ptr(tmp, stream);
							loop = 0;						
						};
					};
				// check if this is a volume id
				}else if((tmp[11]&FAT_VOL_ID)==FAT_VOL_ID){
					
				}
				// check if this is a pointer to another directory
				else if ((tmp[11]&FAT_DIRECTORY)==FAT_DIRECTORY){
					// do want to print ?
					if(mode == FIO_PRINT){
						// this is another directory	
						// USART_sendstring("File",0x01);
						USART_sendpgmstring(0x25,0x00);
						
						// print the filename
						for(i=0; i<11; i++){
								USART_send(tmp[i]);
						};
						USART_sendspaces(20);
						// get the dirposition
						myl = (((u16)tmp[27])<<8) + tmp[26];
						myh = (((u16)tmp[21])<<8) + tmp[20];
						// substract 2 because the address in the fat is always two wrong. this is what ms wants ?!?
						myl -= 2;
						if(myl > 0xfffd){
							myh--;
						};
						USART_sendhex(myh>>8);
						USART_sendhex(myh);
						USART_sendhex(myl>>8);
						USART_sendhex(myl);
						
						USART_newline();
					} else if((mode == FIO_DIRNUM)&&(loop)){
						// is this the file we are looking for ?
						if(index == count){
							// check if we are supposed to use a mask
							if(name) {
								USART_sendstring("cmp : ", 0x00);
								USART_send_mem(name, 11);
								USART_sendstring(" & ", 0x00);
								USART_send_mem(tmp, 11);
								USART_newline();
								if(fio_cmp_names(tmp, name, NULL)){
									fio_load_dir_ptr(tmp, dir);
									loop = 0;
								};
								//index--;
							} else {
								fio_load_dir_ptr(tmp,dir);
								loop = 0;
							};
						} else {
							if(!name){
								index++;
							} else {
								if(fio_cmp_names(tmp, name, NULL)){
									index++;
								};
							}
						};			
					}else if((mode == FIO_DIRNAME)&&(loop)){
						if(fio_cmp_names(tmp, name, NULL)){
							fio_load_dir_ptr(tmp, dir);
							loop = 0;						
						};
					};
				};
				
			};
			// inc loop counter
			len++;
		};
		// stop the reading process+
		MMC_get_sec_stop();
		FAT_get_next_clus_addr(&myhfat, &mylfat);

		//loop = 0;
	} while((myhfat!=0x0fff)&&(mylfat!=0xfffd)&&(loop));
	
	// if we just did some directory listing, we always return true
	if(mode == FIO_DIRNAME) 
		return true;
	
	// we were looking for some file/directory. if loop is true, no data was found
	if(loop) 
		return false;
	
	return true;
};

/*------------------ FIO API LAYER --------------*/

// only the 'r' mode is supported so far
bool fopen(u08 *filename, u08 mode, FILE *stream){
	if(mode=='r'){
		u08 ret = fio_dir_ctl(FIO_FILENAME, stream, NULL, filename, 0);
		stream->file_offset = 0;
		/*
		USART_sendhex(stream->file_current_sectorh>>8);
		USART_sendhex(stream->file_current_sectorh);
		USART_sendhex(stream->file_current_sectorl>>8);
		USART_sendhex(stream->file_current_sectorl);
		*/
		return ret;
	};
	return false;
};

// only the 'r' mode is supported so far
bool fopenc(u08 count, u08 mode, u08 *mask, FILE *stream){
	if(mode=='r'){
		u08 ret;
		ret = fio_dir_ctl(FIO_FILENUM, stream, NULL, mask, count);
		stream->file_offset = 0;
		/*
		USART_sendhex(stream->file_current_sectorh>>8);
		USART_sendhex(stream->file_current_sectorh);
		USART_sendhex(stream->file_current_sectorl>>8);
		USART_sendhex(stream->file_current_sectorl);
		*/
		return ret;
	};
	return false;
};

// closes a filestream this is not really neccesarry, but alot tidier
void fclose(FILE *stream){
	stream->file_first_sectorl = 
	stream->file_first_sectorh = 
	stream->file_current_sectorl = 
	stream->file_current_sectorh =
	stream->file_sizel = 
	stream->file_sizeh = 
	stream->file_offset = NULL;
};



// has the end of the file been reached ?
bool feof(FILE *stream){
	if((stream->file_current_sectorh==0x0fff)&&
	    (stream->file_current_sectorl==0xfffd)&&
		(stream->file_offset == 0x00)
		) return true;
	return false;
};

// checks if a file is open
bool fisopen(FILE *stream){
	return (stream->file_first_sectorl == NULL) &&
			(stream->file_first_sectorh == NULL) &&
			(stream->file_current_sectorl == NULL) && 
			(stream->file_current_sectorh == NULL) &&
			(stream->file_sizel == NULL) && 
			(stream->file_sizeh == NULL) && 
			(stream->file_offset == NULL);
};

// gets data from a file
u08* fget(FILE *stream, u16 size){
	// make sure the FILE is open
	if((stream->file_current_sectorh==0x0)&&(stream->file_current_sectorl==0x0)) return NULL;
	// are we at the end of a sector ?
	if(stream->file_offset == 0x200){
		// load next sector
		stream->file_offset = 0x00;
		FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
		if(feof(stream)) return NULL;
	};	
	// make sure we are not reading in more than the max data buffer
	if(size > FIO_RD_BUF_SIZE) size = FIO_RD_BUF_SIZE;
	// make sure we are not reading past the end of a file
	if((stream->file_sizeh == 0) && (stream->file_sizel < size)) size = stream->file_sizel;
	
	// is all the data in this sector? or do we have to split over two sectors
	if((stream->file_offset+ size) < 0x200){
		// all in one
		fio_get_file_section(stream, size, fio_buf);
		stream->file_offset += size;
	} else {
		// workout where we need to split
		u08 split_pos = 0x200 - stream->file_offset;
		size -= split_pos;
		// get first half
		fio_get_file_section(stream, split_pos, fio_buf);
		// get second half
		FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
		fio_get_file_section(stream, size, &fio_buf[split_pos]);
		stream->file_offset = size;
		size += split_pos;
	};	
	// update the stream struct
	if(stream->file_sizel < size){ 
		// save as we checked for zero earlier
		stream->file_sizeh--;
	};
	stream->file_sizel = size - stream->file_sizel; 
	
	fio_buf[size] = '\0';
	return fio_buf;
};

// this function sets the current position in a file
bool fseek(FILE *stream, u16 posh, u16 posl, u08 whence){
	switch(whence){
	case SEEK_BACK:	
		//set the pos at pos from end
		if(posl > stream->file_sizel ){
			posh++;
		};
		posh = stream->file_sizeh - posh;
		posl = stream->file_sizel - posl;
		// drop through no break; here
	case SEEK_SET:
		// reset size
		stream->file_sizel = stream->file_full_sizel;
		stream->file_sizeh = stream->file_full_sizeh;
		stream->file_offset = 0;
		// make sure we are not reading past the end
		if(((posh == stream->file_full_sizeh) && (posl >= stream->file_full_sizel))
			|| (posh > stream->file_full_sizeh)){
			stream->file_current_sectorh=0x0fff;
			stream->file_current_sectorl=0xfffd;
			return false;
		};
		// reset sectors
		stream->file_current_sectorl = stream->file_first_sectorl;
		stream->file_current_sectorh = stream->file_first_sectorh;
		// loop while we can still get full sectors
		while(!((posh == 0)&&(posl < 0x200))){
			FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
			if(posl<0x200) posh--;
			posl-=0x200;
			if(stream->file_sizel<0x200) stream->file_sizeh--;
			stream->file_sizel-= 0x200;
			
		};
		stream->file_offset = posl;
		break;
	
	case SEEK_CUR:
		// make sure we are not reading past the end
		if(((posh == stream->file_sizeh) && (posl >= stream->file_sizel))
			|| (posh > stream->file_sizeh)){
			stream->file_current_sectorh=0x0fff;
			stream->file_current_sectorl=0xfffd;
			return false;
		};
		// loop while we can still get full sectors
		while(!((posh == 0)&&(posl < 0x200))){
			FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
			if(posl<0x200) posh--;
			posl-=0x200;
			if(stream->file_sizel<0x200) stream->file_sizeh--;
			stream->file_sizel-= 0x200;
		};
		// update offset and make sure, that we did not go past a sector ending
		stream->file_offset += posl;
		if(stream->file_offset > 0x200){
			FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
			stream->file_offset -= 0x200;
		}
		break;
	};	
	/*
	USART_sendhex_16(stream->file_current_sectorh);
	USART_sendhex_16(stream->file_current_sectorl);
	USART_newline();
	
	USART_sendhex_16(stream->file_sizeh);
	USART_sendhex_16(stream->file_sizel);
	USART_newline();
	
	USART_sendhex_16(stream->file_full_sizeh);
	USART_sendhex_16(stream->file_full_sizel);
	USART_newline();
	
	USART_sendhex_16(stream->file_offset);
	USART_newline();
	*/
	
	return true;
};


// plays a sector, by sending it in 32 byte chunks to the mp3 decoder
// returns false if this is the last sector
// else true
// this function does not update the FILE structure
bool fplay_sector(FILE *stream){
	u16 myh = stream->file_current_sectorh;
	u16 myl = stream->file_current_sectorl;
	
	// make sure the FILE is open
	if((stream->file_current_sectorh==0x0)&&(stream->file_current_sectorl==0x0)) return false;
	// make sure, that we are not past the last sector of the file
	if((stream->file_current_sectorh==0x0fff)&&(stream->file_current_sectorl==0xfffd)) return false;
		
	myl += first_data_sectorl;
	if(myl < first_data_sectorl) myh++;
	
	u16 length = 0;
	u08 tmp;
	
	// start getting the data
	MMC_get_sec_start(myh, myl);
	// trigger the 74hc08 so that the MMC can only send and not receive data(gets 0x00)
	NO_ALLOW_SCI;
	tmp = spi_io(0xff);
	while(length< 512){
		if((512%32) == 0) VS1001_WHILE_NEEDS_NO_DATA;
		VS_BSYNC_HI;
		outp(tmp, SPDR);		// send data
		NOP;
		NOP;
		NOP;
		VS_BSYNC_LO;
		// wait for data to be sent
		while(((inp(SPSR))&(1<<SPIF))== 0x00){};
		tmp = inp(SPDR);
		length++;
	};
	// trigger the 74hc08 so that the MMC can receive data
	ALLOW_SCI;
	// get the 2nd CRC bytes
	spi_io(0xff);
	// give enough time
	MMC_cleanup();
	FAT_get_next_clus_addr(&stream->file_current_sectorh, &stream->file_current_sectorl);
	return true;
};

// prints the id3 tag.
// this function is slow as hell. takes around 6 secs for a 4 mb file
// as the uc has to do 8192 32 byte reads plus all the maths ;( 
// this will change when i have a 512 byte buffer as i can read whole sectors from the fat chain
void fid3tag(FILE *stream){
	fseek(stream, 0x00, 0x007d,SEEK_BACK);
	fget(stream,0x1e);
	//USART_sendstring("Title :", 0x01);
	USART_sendpgmstring(0x26,0x01);
	USART_sendstring(fio_buf, 0x01);
	fget(stream,0x1e);
	//USART_sendstring("Artist :", 0x01);
	USART_sendpgmstring(0x27,0x0);
	USART_sendstring(fio_buf, 0x01);
	fget(stream,0x1e);
	//USART_sendstring("Album :", 0x01);
	USART_sendpgmstring(0x28,0x0);
	USART_sendstring(fio_buf, 0x01);
	fget(stream,0x4);
	//USART_sendstring("Year :", 0x01);
	USART_sendpgmstring(0x29,0x0);
	USART_sendstring(fio_buf, 0x01);
	fget(stream,0x1e);
	//USART_sendstring("Comment :", 0x01);
	USART_sendpgmstring(0x2a,0x0);
	USART_sendstring(fio_buf, 0x01);
	fget(stream,0x1);
	//USART_sendstring("Genre   :", 0x01);
	USART_sendpgmstring(0x2b,0x0);
	USART_sendhex(fio_buf[0]);
	USART_newline();
	fseek(stream, 0x00, 0x0000,SEEK_SET);
};


// step into a subfolder of the current folder
// if dirname is null function sets the root folder
bool dir_enter(u08 *dirname, DIR *dir){
	if(!dir) dir = &directory;
	// set a folder 
	if(dirname){
		return fio_dir_ctl(FIO_DIRNAME, NULL, dir, dirname, 0);
	}
	//set root folder ?
	else {
		dir_root(dir);
		return true;
	};
};

// step into a subfolder of the current folder
bool dir_leave(DIR *dir){
	return dir_enter("..      ", dir);
};

// list the current directory
void dir_list(DIR *dir){
	if(!dir) dir = &directory;
	fio_dir_ctl(FIO_PRINT, NULL, NULL, NULL, 0);
};

// sets the dir to root
void dir_root(DIR *dir){
	if(!dir) dir = &directory;
	dir->dir_currentl = 0x00;
	dir->dir_currenth = 0x00;
};

