/*
 * a.lp_mp3 - Open Source Atmel AVR based MP3 Player
 * Copyright (c) 2003-2004 K. John '2B|!2B' Crispin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 *
 * Feedback, Bugs.... mail john{AT}phrozen.org
 *
 */ 

#ifndef __FAT_H__
#define __FAT_H__


//#define FAT_WRITE

/*------------------ FAT LAYER --------------*/

// print a whole sector
void FAT_print_sector(u16 sectorh, u16 sectorl);

//starts the fat
void FAT_startup(void);

// this function gets the addr in the data region for the next cluster of a file
void FAT_get_next_clus_addr(u16* sectorh, u16* sectorl);


/*------------------ FIO LAYER --------------*/
// this datastructutre stores the address of the current directory that we are in
// and the file that we are currently processing
typedef struct _DIR{
	u16 dir_currenth ;		
	u16 dir_currentl;
	u16 dir_current_entry;
} DIR;

// all the info we need for a file
typedef struct _FILE{
	u16 file_current_sectorh;
	u16 file_current_sectorl;
	u16 file_first_sectorh;
	u16 file_first_sectorl;
	u16 file_sizeh;
	u16 file_sizel;
	u16 file_full_sizeh;
	u16 file_full_sizel;
	u16 file_offset;
	u08 name[11];
} FILE;

#ifndef FAT_WRITE
	// this value should never be bigger than 512
	#define FIO_RD_BUF_SIZE 0x20
	// plus one for '\0
	u08 fio_buf[FIO_RD_BUF_SIZE+1];
#endif
#ifdef FAT_WRITE
	// these value should never be bigger than 512
	#define FIO_WR_BUF_SIZE 0x200
	#define FIO_RD_BUF_SIZE FIO_WR_BUF_SIZE
	// plus one for '\0
	u08 fio_buf[FIO_WR_BUF_SIZE+1];
#endif


#define FIO_PRINT  		0x01
#define FIO_FILENAME 	0x02
#define FIO_DIRNAME 	0x03
#define FIO_FILENUM 	0x04
#define FIO_DIRNUM 		0x05
bool fio_dir_ctl(u08 mode, FILE *stream, DIR *dir, u08* name, u08 count);


/*------------------ FIO API LAYER --------------*/

bool fopen(u08 *filename, u08 mode, FILE *stream);
bool fopenc(u08 count, u08 mode, u08* mask, FILE *stream);
void fclose(FILE *stream);
bool feof(FILE *stream);
bool fisopen(FILE *stream);

// this function does not update the FILE structures size parameters
bool fplay_sector(FILE *stream);

u08* fget(FILE *stream, u16 size);

#define SEEK_SET	0x01
#define SEEK_CUR	0x02
#define SEEK_BACK	0x03
bool fseek(FILE *stream, u16 posh, u16 posl, u08 whence);

// slow as hell
void fid3tag(FILE *stream);

// passing NULL as *dir param will make the api use the default internal *directory struct
bool dir_enter(u08 *dirname, DIR *dir);
bool dir_leave(DIR *dir);
void dir_list(DIR *dir);
void dir_root(DIR *dir);

#endif

